#version 450
/*
 * Copyright 2021 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the ECG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */

layout (binding = 0) uniform UniformBuffer {
	vec4 color;
	mat4 modelMatrix;
	mat4 modelMatrixForNormals;
	mat4 viewProjMatrix;
	vec4 cameraPosition;
	vec4 illumination; // ka, kd, ks, alpha
	ivec4 textureIds; // albedo, metallic, roughness, normal
} ub_data;

layout (binding = 1) uniform DirectionalLight {
	vec4 color;
	vec4 direction;
} dl_data;

layout (binding = 2) uniform PointLight {
	vec4 color;
	vec4 position;
	vec4 attenuation;
} pl_data[10];

layout (binding = 3) uniform sampler2D textures[2];

layout (binding = 4) uniform ShaderConstants {
	vec4 illumination_multiplier;
} constants;

layout (location = 0) in VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 textureCoordinates;
} frag_in;

layout (location = 0) out vec4 out_color;

const float PI = 3.14159265359;

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);

    return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}

void main() {

  vec3 albedo = ub_data.color.xyz;
	float metallic = ub_data.illumination[0];
  float roughness = ub_data.illumination[3];
  float ao = 1.0;

	if(ub_data.textureIds[0] >= 0)
	{
		albedo = texture(textures[ub_data.textureIds[0]], frag_in.textureCoordinates).rgb;
		albedo.r = pow(albedo.r, 2.2);
  	albedo.g = pow(albedo.g, 2.2);
		albedo.b = pow(albedo.b, 2.2);
	}
	if(ub_data.textureIds[1] >= 0)
	{
		metallic = texture(textures[ub_data.textureIds[1]], frag_in.textureCoordinates).r;
	}
	if(ub_data.textureIds[2] >= 0)
	{
		roughness = texture(textures[ub_data.textureIds[2]], frag_in.textureCoordinates).r;
	}

	vec3 N = normalize(frag_in.normalWorld);
	vec3 V = normalize(ub_data.cameraPosition.xyz - frag_in.positionWorld);

  vec3 F0 = vec3(0.04);
  F0 = mix(F0, albedo, metallic);

  // reflectance equation
  vec3 Lo = vec3(0.0);

  // for each light source
	for(int i = 0; i < 10; ++i)
	{
		 // calculate per-light radiance
		vec3 L = normalize(pl_data[i].position.xyz - frag_in.positionWorld);
	  vec3 H = normalize(V + L);
	  float distance = length(pl_data[i].position.xyz - frag_in.positionWorld);
	  float attenuation = 1.0/(pl_data[i].attenuation.x + distance * pl_data[i].attenuation.y + distance * distance * pl_data[i].attenuation.z);
	  vec3 radiance = pl_data[i].color.xyz * attenuation;

	  // cook-torrance brdf
	  float NDF = DistributionGGX(N, H, roughness);
	  float G = GeometrySmith(N, V, L, roughness);
	  vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

	  vec3 kS = F;
	  vec3 kD = vec3(1.0) - kS;
	  kD *= 1.0 - metallic;

	  vec3 numerator = NDF * G * F;
	  float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001;
	  vec3 specular = numerator / denominator;

	  // add to outgoing radiance Lo
	  float NdotL = max(dot(N, L), 0.0);
	  Lo += (kD * albedo / PI + specular) * radiance * NdotL;
	}

  // DirectionalLight
  vec3 L = normalize(dl_data.direction.xyz - frag_in.positionWorld);
  vec3 H = normalize(V + L);
  float distance = length(dl_data.direction.xyz - frag_in.positionWorld);
  float attenuation = 1.0;
  vec3 radiance = dl_data.color.xyz * attenuation;

  // cook-torrance brdf
  float NDF = DistributionGGX(N, H, roughness);
  float G = GeometrySmith(N, V, L, roughness);
  vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

  vec3 kS = F;
  vec3 kD = vec3(1.0) - kS;
  kD *= 1.0 - metallic;

  vec3 numerator = NDF * G * F;
  float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001;
  vec3 specular = numerator / denominator;

  // add to outgoing radiance Lo
  float NdotL = max(dot(N, L), 0.0);
  Lo += (kD * albedo / PI + specular) * radiance * NdotL;

  // end DirectionalLight

  // ambient term
  vec3 ambient = vec3(0.03) * albedo * ao;
  vec3 color = ambient + Lo;

  // map HDR and gamma correct
  color = color / (color + vec3(1.0));
  color = pow(color, vec3(1.0/2.2));

	// Write color for the current fragment:
	out_color = vec4(color, 1.0)*constants.illumination_multiplier.x;
	//out_color = vec4(frag_in.normalWorld, 1.0)*constants.illumination_multiplier.x;
}
