#version 450
/*
 * Copyright 2021 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the ECG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */
 
layout (binding = 0) uniform UniformBuffer {
	vec4 color;
	mat4 modelMatrix;
	mat4 modelMatrixForNormals;
	mat4 viewProjMatrix;
	vec4 cameraPosition;
	vec4 illumination; // ka, kd, ks, alpha
} ub_data;

layout (binding = 1) uniform DirectionalLight {
	vec4 color;
	vec4 direction;
} dl_data;

layout (binding = 2) uniform PointLight {
	vec4 color;
	vec4 position;
	vec4 attenuation;
} pl_data;

layout (binding = 3) uniform sampler2D diffuse_texture;

layout (binding = 4) uniform ShaderConstants {
	float illumination_multiplier;
} constants;

layout (location = 0) in VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 textureCoordinates;
} frag_in;

layout (location = 0) out vec4 out_color;

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) {
		att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	}
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

void main() {	
	vec3 n = normalize(frag_in.normalWorld);
	vec3 v = normalize(ub_data.cameraPosition.xyz - frag_in.positionWorld);
	
	// vec3 diffuseColor = texture(diffuse_texture, frag_in.textureCoordinates).rgb;
	vec3 diffuseColor = ub_data.color.rgb;

	// Start with ambient illumination contribution:
	vec3 color = diffuseColor * ub_data.illumination[0];
	
	// Add directional light's contribution:
	color += phong(
		n, 
		-dl_data.direction.xyz, 
		v, 
		dl_data.color.rgb * diffuseColor, ub_data.illumination[1], 
		dl_data.color.rgb,                ub_data.illumination[2], 
		ub_data.illumination[3], 
		false, vec3(1.0)
	);
			
	// Add point light's contribution
	color += phong(
		n, 
		pl_data.position.xyz - frag_in.positionWorld, 
		v, 
		pl_data.color.rgb * diffuseColor, ub_data.illumination[1], 
		pl_data.color.rgb,                ub_data.illumination[2], 
		ub_data.illumination[3], 
		true, pl_data.attenuation.xyz
	);
	
	// Write color for the current fragment:
	out_color = vec4(color, 1.0)*constants.illumination_multiplier;
}

